import { BindingEventService } from '@slickgrid-universal/binding';
import { calculateAvailableSpace, classNameToList, createDomElement, emptyElement, getOffset, getOffsetRelativeToParent, isDefined, } from '@slickgrid-universal/utils';
import { SlickEventHandler } from '../core/index.js';
export class MenuBaseClass {
    /** Constructor of the SlickGrid 3rd party plugin, it can optionally receive options */
    constructor(extensionUtility, pubSubService, sharedService) {
        this.extensionUtility = extensionUtility;
        this.pubSubService = pubSubService;
        this.sharedService = sharedService;
        this._addonOptions = {};
        this._camelPluginName = '';
        this._gridUid = '';
        this._menuCssPrefix = '';
        this._menuPluginCssPrefix = '';
        this.pluginName = '';
        this._bindEventService = new BindingEventService();
        this._eventHandler = new SlickEventHandler();
    }
    get addonOptions() {
        return this._addonOptions;
    }
    set addonOptions(newOptions) {
        this._addonOptions = newOptions;
    }
    get eventHandler() {
        return this._eventHandler;
    }
    get grid() {
        return this.sharedService.slickGrid;
    }
    get gridOptions() {
        return this.sharedService.gridOptions ?? {};
    }
    /** Getter for the grid uid */
    get gridUid() {
        return this._gridUid || (this.grid?.getUID() ?? '');
    }
    get gridUidSelector() {
        return this.gridUid ? `.${this.gridUid}` : '';
    }
    get menuCssClass() {
        return this._menuPluginCssPrefix || this._menuCssPrefix;
    }
    get menuElement() {
        return this._menuElm || document.querySelector(`.${this.menuCssClass}${this.gridUidSelector}`);
    }
    /** Dispose (destroy) of the plugin */
    dispose() {
        this._eventHandler?.unsubscribeAll();
        this._bindEventService.unbindAll();
        this.pubSubService.unsubscribeAll();
        this._commandTitleElm?.remove();
        this._optionTitleElm?.remove();
        this.disposeAllMenus();
        emptyElement(this._menuElm);
        this.menuElement?.remove();
        this._menuElm?.remove();
    }
    /** Remove/dispose all parent menus and any sub-menu(s) */
    disposeAllMenus() {
        this.disposeSubMenus();
        // remove all parent menu listeners before removing them from the DOM
        this._bindEventService.unbindAll('parent-menu');
        document.querySelectorAll(`.${this.menuCssClass}${this.gridUidSelector}`).forEach((subElm) => subElm.remove());
    }
    /**
     * Remove/dispose all previously opened sub-menu(s),
     * it will first remove all sub-menu listeners then remove sub-menus from the DOM
     */
    disposeSubMenus() {
        this._bindEventService.unbindAll('sub-menu');
        document.querySelectorAll(`.${this.menuCssClass}.slick-submenu${this.gridUidSelector}`).forEach((subElm) => subElm.remove());
    }
    setOptions(newOptions) {
        this._addonOptions = { ...this._addonOptions, ...newOptions };
    }
    // --
    // protected functions
    // ------------------
    addSubMenuTitleWhenExists(item, commandOrOptionMenu) {
        if (item !== 'divider' && item?.subMenuTitle) {
            const subMenuTitleElm = document.createElement('div');
            subMenuTitleElm.className = 'slick-menu-title';
            subMenuTitleElm.textContent = item.subMenuTitle;
            const subMenuTitleClass = item.subMenuTitleCssClass;
            if (subMenuTitleClass) {
                subMenuTitleElm.classList.add(...classNameToList(subMenuTitleClass));
            }
            commandOrOptionMenu.appendChild(subMenuTitleElm);
        }
    }
    /** Construct the Command/Options Items section. */
    populateCommandOrOptionItems(itemType, menuOptions, commandOrOptionMenuElm, commandOrOptionItems, args, itemClickCallback, itemMouseoverCallback) {
        if (args && commandOrOptionItems && menuOptions) {
            for (const item of commandOrOptionItems) {
                this.populateSingleCommandOrOptionItem(itemType, menuOptions, commandOrOptionMenuElm, item, args, itemClickCallback, itemMouseoverCallback);
            }
        }
    }
    /** Add the Command/Options Title when necessary. */
    populateCommandOrOptionTitle(itemType, menuOptions, commandOrOptionMenuElm, level) {
        if (menuOptions) {
            const isSubMenu = level > 0;
            // return or create a title container
            const menuHeaderElm = this._menuElm?.querySelector(`.slick-${itemType}-header`) ?? createDomElement('div', { className: `slick-${itemType}-header` });
            // user could pass a title on top of the Commands/Options section
            const titleProp = `${itemType}Title`;
            if (!isSubMenu) {
                if (menuOptions?.[titleProp]) {
                    emptyElement(menuHeaderElm); // make sure title container is empty before adding anything inside it
                    this[`_${itemType}TitleElm`] = createDomElement('span', {
                        className: 'slick-menu-title',
                        textContent: menuOptions[titleProp],
                    });
                    menuHeaderElm.appendChild(this[`_${itemType}TitleElm`]);
                    menuHeaderElm.classList.add('with-title');
                }
                else {
                    menuHeaderElm.classList.add('no-title');
                }
                commandOrOptionMenuElm.appendChild(menuHeaderElm);
            }
        }
    }
    /** Construct the Command/Options Items section. */
    populateSingleCommandOrOptionItem(itemType, menuOptions, commandOrOptionMenuElm, item, args, itemClickCallback, itemMouseoverCallback) {
        let commandLiElm = null;
        if (args && item && menuOptions) {
            const level = args?.level || 0;
            const pluginMiddleName = this._camelPluginName === 'headerButtons' ? '' : '-item';
            const menuCssPrefix = `${this._menuCssPrefix}${pluginMiddleName}`;
            // run each override functions to know if the item is visible and usable
            let isItemVisible = true;
            let isItemUsable = true;
            if (typeof item === 'object') {
                isItemVisible = this.extensionUtility.runOverrideFunctionWhenExists(item.itemVisibilityOverride, args);
                isItemUsable = this.extensionUtility.runOverrideFunctionWhenExists(item.itemUsabilityOverride, args);
            }
            // if the result is not visible then there's no need to go further
            if (!isItemVisible) {
                return null;
            }
            // when the override is defined (and previously executed), we need to use its result to update the disabled property
            // so that "handleMenuItemCommandClick" has the correct flag and won't trigger a command/option clicked event
            if (typeof item === 'object' && item.itemUsabilityOverride) {
                item.disabled = isItemUsable ? false : true;
            }
            commandLiElm = createDomElement('li', { className: menuCssPrefix, role: 'menuitem' });
            if (typeof item === 'object' && isDefined(item[itemType])) {
                commandLiElm.dataset[itemType] = item?.[itemType];
            }
            if (commandOrOptionMenuElm) {
                commandOrOptionMenuElm.appendChild(commandLiElm);
            }
            if ((typeof item === 'object' && item.divider) || item === 'divider') {
                commandLiElm.classList.add(`${menuCssPrefix}-divider`);
                return commandLiElm;
            }
            if (item.disabled) {
                commandLiElm.classList.add(`${menuCssPrefix}-disabled`);
            }
            if (item.hidden || item.showOnHover) {
                commandLiElm.classList.add(`${menuCssPrefix}-hidden`);
            }
            if (item.cssClass) {
                commandLiElm.classList.add(...classNameToList(item.cssClass));
            }
            if (item.tooltip) {
                commandLiElm.title = item.tooltip;
            }
            if (this._camelPluginName !== 'headerButtons') {
                // Menu plugin can use optional icon & content elements
                const iconElm = createDomElement('div', { className: `${this._menuCssPrefix}-icon` });
                commandLiElm.appendChild(iconElm);
                if (item.iconCssClass) {
                    iconElm.classList.add(...classNameToList(item.iconCssClass));
                }
                else if (!item.commandItems && !item.optionItems) {
                    iconElm.textContent = '◦';
                }
                const textElm = createDomElement('span', {
                    className: `${this._menuCssPrefix}-content`,
                    textContent: (typeof item === 'object' && item.title) || '',
                }, commandLiElm);
                if (item.textCssClass) {
                    textElm.classList.add(...classNameToList(item.textCssClass));
                }
            }
            // execute command callback on menu item clicked
            const eventGroupName = level > 0 ? 'sub-menu' : 'parent-menu';
            this._bindEventService.bind(commandLiElm, 'click', ((e) => itemClickCallback.call(this, e, itemType, item, level, args?.column)), undefined, eventGroupName);
            // optionally open sub-menu(s) by mouseover
            if (this._addonOptions?.subMenuOpenByEvent === 'mouseover' &&
                typeof itemMouseoverCallback === 'function') {
                this._bindEventService.bind(commandLiElm, 'mouseover', ((e) => itemMouseoverCallback.call(this, e, itemType, item, level, args?.column)), undefined, eventGroupName);
            }
            // the option/command item could be a sub-menu if it has another list of commands/options
            if (item.commandItems || item.optionItems) {
                const chevronElm = document.createElement('span');
                chevronElm.className = 'sub-item-chevron';
                if (this._addonOptions.subItemChevronClass) {
                    chevronElm.classList.add(...classNameToList(this._addonOptions.subItemChevronClass));
                }
                else {
                    chevronElm.textContent = '⮞'; // ⮞ or ▸
                }
                commandLiElm.classList.add('slick-submenu-item');
                commandLiElm.appendChild(chevronElm);
            }
        }
        return commandLiElm;
    }
    /**
     * Reposition any of the menu plugins (CellMenu, ContextMenu, GridMenu, HeaderMenu) to where the user clicked,
     * it will calculate the best position depending on available space in the viewport and the menu type.
     */
    repositionMenu(e, menuElm, buttonElm, addonOptions) {
        const targetElm = e.target; // get header button createElement
        const targetEvent = e?.touches?.[0] ?? e;
        const isSubMenu = menuElm.classList.contains('slick-submenu');
        const rowHeight = this.gridOptions.rowHeight || 0;
        const parentElm = isSubMenu
            ? e.target.closest('.slick-menu-item')
            : this.pluginName === 'CellMenu' || this.pluginName === 'ContextMenu'
                ? e.target.closest('.slick-cell')
                : targetEvent.target;
        if (menuElm && parentElm) {
            // for Cell/Context Menus we should move to (0,0) coordinates before calculating height/width
            // since it could end up being cropped width values when element is outside browser viewport.
            if (this.pluginName === 'CellMenu' || this.pluginName === 'ContextMenu') {
                menuElm.style.top = `0px`;
                menuElm.style.left = `0px`;
            }
            const containerElm = this.sharedService.gridContainerElement.classList.contains('slickgrid-container')
                ? this.sharedService.gridContainerElement
                : (this.sharedService.gridContainerElement.querySelector('.slickgrid-container') ?? this.sharedService.gridContainerElement);
            const relativePos = getOffsetRelativeToParent(containerElm, targetElm);
            const menuWidth = menuElm.offsetWidth;
            const parentOffset = getOffset(parentElm);
            let menuOffsetLeft = 0;
            let menuOffsetTop = 0;
            let dropOffset = 0;
            let sideOffset = 0;
            let availableSpaceBottom = 0;
            let availableSpaceTop = 0;
            const { bottom: parentSpaceBottom, top: parentSpaceTop } = calculateAvailableSpace(parentElm);
            if (this.pluginName === 'GridMenu' && buttonElm) {
                if (!isSubMenu) {
                    const buttonComptStyle = getComputedStyle(buttonElm);
                    const buttonWidth = parseInt(buttonComptStyle?.width ?? addonOptions?.menuWidth, 10);
                    const contentMinWidth = addonOptions?.contentMinWidth ?? 0;
                    const currentMenuWidth = (contentMinWidth > menuWidth ? contentMinWidth : menuWidth) || 0;
                    if (contentMinWidth > 0) {
                        menuElm.style.minWidth = `${contentMinWidth}px`;
                    }
                    const menuIconOffset = getOffset(buttonElm); // get button offset position
                    const nextPositionLeft = menuIconOffset.right;
                    menuOffsetTop = menuIconOffset.top + buttonElm.offsetHeight; // top position has to include button height so the menu is placed just below it
                    menuOffsetLeft =
                        addonOptions?.dropSide === 'right' ? nextPositionLeft - buttonWidth : nextPositionLeft - currentMenuWidth;
                }
            }
            else if (this.pluginName === 'CellMenu' || this.pluginName === 'ContextMenu') {
                menuOffsetLeft = parentElm && this.pluginName === 'CellMenu' ? parentOffset.left : targetEvent.pageX;
                menuOffsetTop = parentElm && this.pluginName === 'CellMenu' ? parentOffset.top : targetEvent.pageY;
                dropOffset = Number(addonOptions?.autoAdjustDropOffset || 0);
                sideOffset = Number(addonOptions?.autoAlignSideOffset || 0);
            }
            else {
                menuOffsetLeft = isSubMenu ? parentOffset.left : (relativePos?.left ?? 0);
                menuOffsetTop = isSubMenu
                    ? parentOffset.top
                    : (relativePos?.top ?? 0) + (addonOptions?.menuOffsetTop ?? 0) + targetElm.clientHeight;
            }
            if ((this.pluginName === 'ContextMenu' || this.pluginName === 'GridMenu') && isSubMenu) {
                menuOffsetLeft = parentOffset.left;
                menuOffsetTop = parentOffset.top;
            }
            // for sub-menus only, auto-adjust drop position (up/down)
            // we first need to see what position the drop will be located (defaults to bottom)
            // since we reposition menu below slick cell, we need to take it in consideration and do our calculation from that element
            const menuHeight = menuElm?.offsetHeight || 0;
            if ((this.pluginName === 'GridMenu' || this.pluginName === 'HeaderMenu') && isSubMenu) {
                availableSpaceBottom = parentSpaceBottom;
                availableSpaceTop = parentSpaceTop;
            }
            else if (this.pluginName === 'CellMenu' ||
                this.pluginName === 'ContextMenu' ||
                addonOptions?.autoAdjustDrop ||
                addonOptions?.dropDirection) {
                availableSpaceBottom = parentSpaceBottom + dropOffset - rowHeight;
                availableSpaceTop = parentSpaceTop - dropOffset + rowHeight;
            }
            const dropPosition = availableSpaceBottom < menuHeight && availableSpaceTop > availableSpaceBottom ? 'top' : 'bottom';
            if (dropPosition === 'top' || addonOptions?.dropDirection === 'top') {
                menuElm.classList.remove('dropdown');
                menuElm.classList.add('dropup');
                if (isSubMenu) {
                    menuOffsetTop -= menuHeight - dropOffset - parentElm.clientHeight;
                }
                else {
                    menuOffsetTop -= menuHeight - dropOffset;
                }
            }
            else {
                menuElm.classList.remove('dropup');
                menuElm.classList.add('dropdown');
                if (this.pluginName === 'CellMenu' || this.pluginName === 'ContextMenu') {
                    menuOffsetTop = menuOffsetTop + dropOffset;
                    if (this.pluginName === 'CellMenu') {
                        if (isSubMenu) {
                            menuOffsetTop += dropOffset;
                        }
                        else {
                            menuOffsetTop += rowHeight + dropOffset;
                        }
                    }
                }
            }
            // when auto-align is set, it will calculate whether it has enough space in the viewport to show the drop menu on the right (default)
            // if there isn't enough space on the right, it will automatically align the drop menu to the left
            // to simulate an align left, we actually need to know the width of the drop menu
            if (addonOptions?.autoAlign ||
                addonOptions?.autoAlignSide ||
                addonOptions?.dropSide === 'left') {
                let subMenuPosCalc = menuOffsetLeft + Number(menuWidth); // calculate coordinate at caller element far right
                if (isSubMenu) {
                    subMenuPosCalc += parentElm.clientWidth;
                }
                const gridPos = this.grid.getGridPosition();
                const browserWidth = document.documentElement.clientWidth;
                const dropSide = subMenuPosCalc >= gridPos.width || subMenuPosCalc >= browserWidth ? 'left' : 'right';
                let needHeaderMenuOffsetLeftRecalc = false;
                if (dropSide === 'left' || (!isSubMenu && addonOptions?.dropSide === 'left')) {
                    menuElm.classList.remove('dropright');
                    menuElm.classList.add('dropleft');
                    if (this.pluginName === 'HeaderMenu') {
                        if (isSubMenu) {
                            menuOffsetLeft -= menuWidth;
                        }
                        else {
                            needHeaderMenuOffsetLeftRecalc = true;
                        }
                    }
                    else if (this.pluginName === 'CellMenu' && !isSubMenu) {
                        const parentCellWidth = parentElm.offsetWidth || 0;
                        menuOffsetLeft -= Number(menuWidth) - parentCellWidth - sideOffset;
                    }
                    else if (this.pluginName !== 'GridMenu' || (this.pluginName === 'GridMenu' && isSubMenu)) {
                        menuOffsetLeft -= Number(menuWidth) - sideOffset;
                    }
                }
                else {
                    menuElm.classList.remove('dropleft');
                    menuElm.classList.add('dropright');
                    if (isSubMenu) {
                        menuOffsetLeft += sideOffset + parentElm.offsetWidth;
                    }
                    else {
                        if (this.pluginName === 'HeaderMenu') {
                            needHeaderMenuOffsetLeftRecalc = true;
                        }
                        else {
                            menuOffsetLeft += sideOffset;
                        }
                    }
                }
                if (needHeaderMenuOffsetLeftRecalc) {
                    menuOffsetLeft = relativePos?.left ?? 0;
                    if (addonOptions?.autoAlign && gridPos?.width && menuOffsetLeft + (menuElm.clientWidth ?? 0) >= gridPos.width) {
                        menuOffsetLeft =
                            menuOffsetLeft + targetElm.clientWidth - menuElm.clientWidth + (addonOptions?.autoAlignOffset || 0);
                    }
                }
            }
            // ready to reposition the menu
            menuElm.style.top = `${menuOffsetTop}px`;
            menuElm.style.left = `${menuOffsetLeft}px`;
            if (this.pluginName === 'GridMenu') {
                menuElm.style.opacity = '1';
                menuElm.style.display = 'block';
            }
        }
    }
}
//# sourceMappingURL=menuBaseClass.js.map